--[[
**********************************************************************
MagicDKP Database - Stores DKP data imported via eqDKP Plus getdkp.php
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]


local MODULE_NAME = "Database"
local mod = MagicDKP
local module = mod:NewModule(MODULE_NAME, "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)
local BC = LibStub("LibBabble-Class-3.0") :GetUnstrictLookupTable()
local WHISPER_PREFIX = "<D> "
local db
local debug
local dkpdb
--[===[@debug@
debug = true
--@end-debug@]===]
if debug then
   MDB = module
end
module.defaults = {
   clearGlobals = true,
   filterZero = true, 
   filterRank = true,
   filterGuild = true,
   filterWhispers = true, 
   enableWhisperQuery = true,
   whisperKeyword = "dkp",
}

local classes = {
   ["death knight"] = BC["DEATHKNIGHT"] or "death knight", 
   ["deathknight"] = BC["DEATHKNIGHT"] or "death knight", 
   druid   = BC.Druid   or "druid",
   hunter  = BC.Hunter  or "hunter",
   mage    = BC.Mage    or "mage", 
   paladin = BC.Paladin or "paladin",
   priest  = BC.Priest  or "priest",
   rogue   = BC.Rogue   or "rogue",
   shaman  = BC.Shaman  or "shaman", 
   warlock = BC.Warlock or "warlock",
   warrior = BC.Warrior or "warrior"
}

local function clear(data) for id in pairs(data) do data[id] = nil end return data end
   
function module:OnInitialize()
   module.guildMembers = {}
   module.multiColumns = { }
   module.multiPools = {}

   -- Set up reverse lookup if needed
   for val, key in pairs(classes) do
      key = key:lower()
      if not classes[key] then classes[key] = val end
   end
   
   dkpdb = MagicDKP_Database or {
      dkp = { },
      multidkp = {
	 dkp = {
	    desc = "Raid DKP",
	    events = "All Events", 
	 }
      },
      lastImportDate = 0
   }
   MagicDKP_Database = dkpdb
   
   db = mod:GetModuleDatabase(MODULE_NAME, module)

   if DKPInfo then
      module:ImportGetDKPData()
   end
   if db.clearGlobals then
      multiTable = nil
      DKPInfo = nil
      gdkp = nil
      DKP_ITEMS = nil
      getdkp_itemids = nil
      gdkp_alliases = nil
      GetDKPRaidPlaner = nil
   end
   for key, data in pairs(dkpdb.multidkp) do
      module.multiPools[key] = data.desc
      module.multiColumns[key] = {
	 earned = key.."_earned", 
	 spent = key.."_spend",
	 adjust = key.."_adjust",
	 current = key.."_current", 
      }
   end

   for player, data in pairs(dkpdb.dkp) do
      classes[data.class:lower()] = data.class
   end
end


function module:HandleChatEvents(disable)
   if db.enableWhisperQuery and not disable then
      module:RegisterEvent("CHAT_MSG_WHISPER", "HandleWhisperQuery")
   else
      module:UnregisterEvent("CHAT_MSG_WHISPER")
   end
   if db.enableWhisperQuery and db.filterWhispers and not disable then
      ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", module.WhisperFilterIn)
      ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", module.WhisperFilterOut)
   else
      ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER", module.WhisperFilterIn)
      ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", module.WhisperFilterOut)
   end
end

function module:OnEnable()
   module:RegisterEvent("GUILD_ROSTER_UPDATE")
   GuildRoster()
   module:HandleChatEvents()
end

function module:OnDisable()
   module:HandleChatEvents(true)
end

function module:GetMultiDKPPools()
   return module.multiPools
end
				

function module:GetMultiDKP()
   return module.multiColumns
end

function module:GetMultiCols(multi)
   local data
   if multi then
      data =  module.multiColumns[multi]
   end
   if not data then
      data = module.multiColumns[db.defaultMulti]
	 or module.multiColumns.dkp
	 or (#dkpdb.multidkp > 0 and module.multiColumns[(next(dkpdb.multidkp))])
   end
   return data
end
   
function module:GetPlayerData(player)
   return dkpdb.dkp[player]
end

function module:GetPlayerItems(player)
   local data = dkpdb.dkp[player] and dkpdb.dkp[player].items
   if not data or #data == 0 then return end
   
   table.sort(data, function(a, b) return (a.dkp or 0) > (b.dkp or 0) end)
   return data
end

function module:AdjustSpentDKP(player, change, multi)
   local cols = module:GetMultiCols(multi)
   local dkp = dkpdb.dkp[player] or {}
   if not cols then return end
   if mod.debug then mod:debug("Player %s has spent %d dkp from %s", player, change, cols.spent) end
   dkp[cols.current] = (dkp[cols.current] or 0) - change
   dkp[cols.spent]   = (dkp[cols.spent] or 0) + change
   dkp[cols.earned]  = dkp[cols.earned] or 0
   dkp[cols.adjust]  = dkp[cols.adjust] or 0
   return dkp
end

function module:AdjustEarnedDKP(player, change, multi)
   local cols = module:GetMultiCols(multi)
   local dkp = dkpdb.dkp[player] or {}
   if not cols then return end 
   dkp[cols.current] = (dkp[cols.current] or 0) + change
   dkp[cols.earned]  = (dkp[cols.earned]  or 0) + change
   dkp[cols.spent]   = dkp[cols.spent] or 0
   dkp[cols.adjust]  = dkp[cols.adjust]  or 0
   dkpdb.dkp[player] = dkp
   return dkp
end   


function module:GetPlayerDKP(player, multi)
   local dkp = dkpdb.dkp[player]
   local cols = module:GetMultiCols(multi)
   if dkp and cols then
      return dkp[cols.current],  dkp[cols.earned],  dkp[cols.spent],  dkp[cols.adjust]
   end
end

function module:GetItemBidSummary(item)
   local min, max, count
   local total = 0
   local name = GetItemInfo(item)
   if type(item) == "string" then
      item = mod:LinkToItemID(item)
   end
   for player, data in pairs(dkpdb.dkp) do
      if data.items then
	 for _,item in pairs(data.items) do
	    if item.itemid == item or item.name == name  then
	       if not min or item.dkp < min then
		  min = item.dkp
	       end
	       if not max or item.dkp > max then
		  max = item.dkp
	       end
	       total = total + item.dkp
	       count = (count or 0) + 1
	    end
	 end
      end
   end

   if count then return min, max, math.floor(total/count), count end
end

function module:GetItemBidHistory(item)
   local result = {}
   local name = GetItemInfo(item)
   if type(item) == "string" then
      item = mod:LinkToItemID(item)
   end
   for player, data in pairs(dkpdb.dkp) do
      if data.items then
	 for _,item in pairs(data.items) do
	    if item.itemid == item or item.name == name  then
	       result[#result+1] = {
		  name = player,
		  dkp = item.dkp,
	       }
	    end
	 end
      end
   end
   table.sort(result, function(a, b) return (a.dkp or 0) > (b.dkp or 0) end)
   return result
end


-- Import data from the GetDKP file
function module:ImportGetDKPData()
   local isNew = tonumber(DKPInfo.timestamp) > dkpdb.lastImportDate
   if not isNew then return end
   
   dkpdb.dkp = {}
   if mod.hasDebug then
      mod:debug("Importing player dkp...")
   end
   for name, data in pairs(gdkp.players) do
      if not db.filterZero or (data.DKP or data.dkp_current) > 0 then
	 local p = {}
	 for var, val in pairs(data) do
	    if var ~= "rcount" then
	       p[var] = val
	       if mod.hasSpam then
		  mod:spam("   %s = %s", var, tostring(val))
	       end
	    end
	 end
	 if mod.nameSubstitution then
	    name = mod.nameSubstitution[name] or name
	 end
	 dkpdb.dkp[name] = p
      end
   end
   if DKP_ITEMS then
      if mod.hasDebug then mod:debug("Importing items...") end
      for name, data in pairs(DKP_ITEMS)  do
	 if mod.nameSubstitution then
	    name = mod.nameSubstitution[name] or name
	 end
	 local p = dkpdb.dkp[name]
	 if p then
	    p.items = {}
	    for id,item in pairs(data.Items) do
	       p.items[id] = {
		  dkp = item.dkp,
		  name = item.name,
		  itemid = getdkp_itemids and getdkp_itemids[item.name]
	       }
	       if mod.hasSpam then
		  mod:spam("   %s  (%s)", item.name, tostring(p.items[id].itemid))
	       end	       
	    end
	 end
      end
   end

   dkpdb.multidkp = {}
   if multiTable then
      for _, data in ipairs(multiTable) do 
	 for key, val in pairs(data) do
	    dkpdb.multidkp[key] = {
	       desc = val.disc,
	       events = val.events
	    }
	 end
      end
   end
   dkpdb.lastImportDate = tonumber(DKPInfo.timestamp)
   module.ImportGetDKPData = nil
end

local function cap(str)
   return str:lower():trim():gsub("^.", string.upper)
end

local function findpattern(text, pattern)
   local a, b, tail = string.find(text, pattern)
   if a and b then 
      return (string.sub(text, a, b)), tail and tail:trim()
   end
end

local function findwhispercmd(msg)
   if not module.whisperPattern then
      module.whisperPattern = "^"..db.whisperKeyword.." (.*)"
   end
   msg = msg:trim()
   if msg == db.whisperKeyword then
      return msg, nil, msg
   else
      local a, b =  findpattern(msg, module.whisperPattern)
      return a, b, msg
   end
   return nil, nil, msg
end

local function cap(str)
   return str:lower():trim():gsub("^.", string.upper)
end

function module:HandleWhisperQuery(channel, msg, sender)
   local cmd, param 
   cmd, param, msg = findwhispercmd(msg)
   if not cmd then return end
   if param and cap(param) ~= sender then
      local itemid = mod:LinkToItemID(param)
      if itemid then
	 module:WhisperItemSummary(itemid, sender)
      else 
	 local lp = param:lower()
	 if lp == "help" then
	    module:WhisperHelp(sender)
	 elseif lp == "tier" then
	 elseif lp == "class" then
	    local pl = UnitClass(sender)
	    if pl then
	       module:WhisperClassData((UnitClass(sender)), sender)
	    else
	       mod:SendChatMessage(L["Can't determine your class. Please query with an explicit class instead: %s <class name>"]:format(db.whisperKeyword), "WHISPER", sender, WHISPER_PREFIX)
	    end
	 elseif classes[lp] then
	    module:WhisperClassData(param, sender)
	 else
	    module:WhisperPlayerData(cap(param), sender)
	 end
      end
   else
      module:WhisperPlayerData(sender)
   end
end

if GetNumTalentGroups then
   -- 3.1 support
   function module:WhisperFilterOut(event, msg, ...)
      return string.find(msg, WHISPER_PREFIX) == 1, msg, ...
   end
   function module:WhisperFilterIn(event, msg, ...)
      return (findwhispercmd(msg))
   end
else
   function module.WhisperFilterIn(msg)
      return (findwhispercmd(msg)), msg
   end
   function module.WhisperFilterOut(msg)
      return string.find(msg, WHISPER_PREFIX) == 1
   end
end

function module:WhisperItemSummary(itemid, recipient)
   local min, max, avg, count = module:GetItemBidSummary(itemid)
   local link = (select(2, GetItemInfo(itemid)))
   if count then
      mod:SendChatMessage(L["Min Bid: %d  Max Bid: %d  Average Bid: %d  Total Drops: %d."]:format(min, max, avg, count), "WHISPER", recipient, WHISPER_PREFIX)
   else
      mod:SendChatMessage(L["No bid information found for %s."]:format(link), "WHISPER", recipient, WHISPER_PREFIX)
   end
end

function module:WhisperHelp(recipient)
   mod:SendChatMessage(L["Available whisper commands:"], "WHISPER", recipient, WHISPER_PREFIX)
   mod:SendChatMessage(L["Get own DKP: %s"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
   mod:SendChatMessage(L["Get player DKP: %s <player name>"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
   mod:SendChatMessage(L["Get your own class DKP: %s class"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
   mod:SendChatMessage(L["Get specific class DKP: %s <class name>"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
--   mod:SendChatMessage(L["Get tier token DKP: %s tier"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
   mod:SendChatMessage(L["Get item bid summary: %s [item link]"]:format(db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
end

do
   local response = {}
   local classData = {}
   function module:WhisperPlayerData(player, recipient, short)
      local data = module:GetPlayerData(player)
      recipient = recipient or player
      if not data then
	 local alt = mod:GetPlayerMain(player)
	 if alt then
	    module:WhisperPlayerData(alt, recipient)
	    return
	 end
	 mod:SendChatMessage(L["No DKP data found for %s. Try \"/w %s %s help\" for more information."]:format(player, UnitName("player"), db.whisperKeyword), "WHISPER", recipient, WHISPER_PREFIX)
	 return
      end
      clear(response)
      for name, cols in pairs(module.multiColumns) do
	 local dkp = data[cols.current]
	 if dkp and dkp > 0 then
	    response[#response+1] = name.." ".. dkp
	 end
      end
      if short then
	 mod:SendChatMessage(player ..": "..table.concat(response, ", "), "WHISPER", recipient, WHISPER_PREFIX)
      else
	 mod:SendChatMessage(L["Current DKP for %s: %s"]:format(player, table.concat(response, ", ")), "WHISPER", recipient, WHISPER_PREFIX)
      end
   end

   function module:ShowPlayerWhisper(id, recipient)
      if id == recipient then return true end
      if not next(module.guildMembers) then return true end
      local rank = module.guildMembers[id]
      if (db.filterGuild and not rank) or
	 (db.filterRank and (not rank or not mod:IsRaiderRank(rank))) then
	 return false
      end
      return true
   end

   function module:WhisperClassData(class, recipient)
      clear(response)
      clear(classData)
      local lc = class:lower()
      local bc = classes[lc]:lower()
      local pdb = dkpdb.dkp
      for id, data in pairs(pdb) do
	 local pc = data.class:lower()
	 if (pc == lc or pc == bc) and
	    module:ShowPlayerWhisper(id, recipient) then
	    classData[#classData+1] = id
	 end
      end
      if next(classData) then
	 table.sort(classData, function(a, b) return (pdb[a].DKP or pdb[a].dkp_current or 0) > (pdb[b].DKP or pdb[b].dkp_current or 0) end)
	 mod:SendChatMessage(L["Current DKP for %s:"]:format(class), "WHISPER", recipient, WHISPER_PREFIX)
	 for _, player in ipairs(classData) do
	    module:WhisperPlayerData(player, recipient, true)
	 end
      else
	 mod:SendChatMessage(L["No DKP data found for:  "]..class, "WHISPER", recipient, WHISPER_PREFIX)
      end
   end
end

function module:GUILD_ROSTER_UPDATE(...)
   module:UnregisterEvent("GUILD_ROSTER_UPDATE") 
   clear(module.guildMembers)
   local numGuildMembers = GetNumGuildMembers(true)
   for i = 1, numGuildMembers do
      local name, _, rankIndex = GetGuildRosterInfo(i)
      module.guildMembers[name] = rankIndex
   end
end

function module:SetProfileParam(var, value, sub)
   local varName = var[#var]
   if sub ~= nil then
      if type(db[varName]) ~= "table" then db[varName] = mod.get() end
      db[varName][value] = sub
   else
      db[varName] = value      
   end
   module:HandleChatEvents()
   module.whisperPattern = nil
end

function module:GetProfileParam(var, sub)
   local varName = var[#var]
   if sub ~= nil then
      return type(db[varName]) == "table" and db[varName][sub]
   else
      return db[varName]
   end
end

function module:OnProfileChanged(newdb)
   db = newdb
end

module.options = {
   type = "group",
   name = L["DKP Database"],
   handler = module,
   childGroups  = "tab",
   set = "SetProfileParam",
   get = "GetProfileParam",
   args = {
      defaultMulti  = {
	 width = "full", 
	 type = "select",
	 name = L["Default DKP Pool"],
	 desc = L["The default DKP to use when logging and fetching DKP. This will be used when bidding using MagicDKP Bidder and when recording DKP changes. Zone and Item level settings in MagicDKP can be used to override the default."],
	 values = "GetMultiDKPPools",
	 order = 1, 
      },      
      clearGlobals = {
	 type = "toggle",
	 width = "full", 
	 name = L["Clear Globals After Import"],
	 desc = L["Clear the global variables used in the data import after import is done. This option will save memory by removing unused data. Do not enable if you has another addon enabled that uses this data."], 
      },
      filterZero = {
	 type = "toggle",
	 width = "full", 
	 name = L["Filter Zero DKP Entries"],
	 desc = L["Don't import any information for players that has zero DKP earned."], 
      },
      filterGuild = {
	 type = "toggle",
	 width = "full", 
	 name = L["Filter Guild Only entries"],
	 desc = L["Don't show players that is not in the guild when answering dkp whispers."], 
      },
      filterRank = {
	 type = "toggle",
	 width = "full", 
	 name = L["Filter by Rank"],
	 desc = L["Filter out non-raiding ranks in whisper responses."], 
      },
      filterWhispers = {
	 type = "toggle",
	 width = "full", 
	 name = L["Filter DKP Whispers"],
	 desc = L["Filter in and outgoing DKP whisper commands."], 
      },
      enableWhisperQuery = {
	 type = "toggle",
	 width = "full",
	 name = L["Enable Whisper Queries"],
	 desc = L["Allow players to whisper you a query to check their or a classes DKP status."], 
      },
      whisperKeyword = {
	 type = "input",
	 name = L["Whisper Keyword"],
	 desc = L["The whisper command used to query DKP status."], 
      },
      
   }
}
